from __future__ import annotations

import asyncio
import json
import logging
import uuid
from dataclasses import dataclass
from typing import Literal, get_args
from urllib.parse import urlparse

import aiohttp
from PySide2.QtCore import QObject, Signal, Slot

from .async_worker import AsyncWorker

_logger = logging.getLogger(__name__)

HttpMethod = Literal["GET", "POST", "PUT", "DELETE", "PATCH", "HEAD", "OPTIONS"]
ResponseType = Literal["text", "json", "bytes", "status"]


@dataclass(frozen=True)
class RequestResult:
    """Represents the successful outcome of an HTTP request."""
    request_id: uuid.UUID
    data: str | dict | list | bytes | int
    status_code: int
    headers: dict[str, str]


@dataclass(frozen=True)
class RequestError:
    """Contains detailed information about a failed HTTP request."""
    request_id: uuid.UUID
    error_type: str
    message: str
    status_code: int | None = None
    headers: dict[str, str] | None = None
    raw_response: str | bytes | None = None


class AsyncHttpClient(QObject):
    request_started = Signal(uuid.UUID)
    request_completed = Signal(RequestResult)
    request_cancelled = Signal(uuid.UUID)
    request_failed = Signal(RequestError)

    def __init__(
            self,
            max_concurrent: int = 10,
            timeout: float = 30,
            timeout_connect: float = 10,
            limit: int = 100,
            limit_per_host: int = 10,
            max_retries: int = 3,
            retry_backoff: float = 1.0,
            verify_ssl: bool = True,
            parent: QObject | None = None):
        super().__init__(parent)
        self._max_concurrent = max_concurrent
        self._timeout = timeout
        self._timeout_connect = timeout_connect
        self._limit = limit
        self._limit_per_host = limit_per_host
        self._verify_ssl = verify_ssl
        self._max_retries = max_retries
        self._retry_backoff = retry_backoff

        self._session: aiohttp.ClientSession | None = None
        self._worker: AsyncWorker | None = None

    def send_request(
            self,
            url: str,
            method: HttpMethod = "GET",
            *,
            params: dict[str, str] | None = None,
            headers: dict[str, str] | None = None,
            response_type: ResponseType = "text"
    ) -> uuid.UUID:
        """Submit a new HTTP request for asynchronous processing."""
        try:
            parsed = urlparse(url)
            if not all([parsed.scheme, parsed.netloc]):
                raise ValueError("Invalid URL format")
            if parsed.scheme not in ('http', 'https'):
                raise ValueError("Unsupported URL scheme")
        except Exception as e:
            raise ValueError(f"Invalid URL: {str(e)}") from e
        if method not in get_args(HttpMethod):
            raise ValueError(f"Invalid HTTP method: {method}")
        if response_type not in get_args(ResponseType):
            raise ValueError(f"Invalid response type: {response_type}")

        if (self._worker and
                self._worker.get_active_task_count() >= self._max_concurrent):
            raise RuntimeError(f"Maximum concurrent requests ({self._max_concurrent}) reached")

        if not self._worker or not self._worker.is_running():
            self._worker = AsyncWorker()

        return self._worker.run_task(
            self._execute_http_request,
            url,
            method,
            params=params,
            headers=headers,
            response_type=response_type
        )

    @Slot(uuid.UUID)
    def cancel_request(self, request_id: uuid.UUID) -> None:
        """Cancel a pending request."""
        if not self.is_running():
            return

        self._worker.cancel_task(request_id)

    @Slot()
    def shutdown(self) -> None:
        """Immediately shutdown client and release resources."""
        if self.is_running():
            self._worker.stop(self._close_session, self._session)

    def is_running(self) -> bool:
        return self._worker and self._worker.is_running()

    def has_active_requests(self) -> bool:
        if not self.is_running():
            return False
        return self._worker.get_active_task_count() > 0

    async def _execute_http_request(
            self,
            task_id: uuid.UUID,
            url: str,
            method: HttpMethod = "GET",
            *,
            params: dict[str, str] | None = None,
            headers: dict[str, str] | None = None,
            response_type: ResponseType = "text"
    ) -> None:
        """Execute HTTP request with advanced retry logic."""
        try:
            # Ensure we have an active session
            await self._ensure_session()
            self.request_started.emit(task_id)

            last_status = None

            for attempt in range(self._max_retries + 1):
                try:
                    async with self._session.request(
                            method,
                            url,
                            headers=headers,
                            params=params,
                            raise_for_status=False
                    ) as response:
                        last_status = response.status

                        # Successful responses (2xx)
                        if 200 <= response.status < 300:
                            content = await self._process_response(response, response_type)
                            result = RequestResult(
                                request_id=task_id,
                                data=content,
                                status_code=response.status,
                                headers=dict(response.headers)
                            )
                            self.request_completed.emit(result)
                            return

                        # Determine if we should retry the request
                        should_retry = False
                        retry_delay = self._retry_backoff * (2 ** attempt)

                        # Special handling for retryable status codes
                        if response.status == 429:  # Too Many Requests
                            should_retry = attempt < self._max_retries
                            retry_after = response.headers.get('Retry-After')
                            if retry_after:
                                try:
                                    retry_delay = float(retry_after)
                                except ValueError:
                                    # If Retry-After contains a date instead of seconds
                                    pass

                        elif response.status == 503:  # Service Unavailable
                            should_retry = attempt < self._max_retries
                            retry_delay = min(retry_delay * 2, 60)  # Max 60 seconds

                        elif 500 <= response.status < 600:  # Server errors
                            should_retry = attempt < self._max_retries

                        # If we shouldn't retry, send error response
                        if not should_retry:
                            raw_response = await response.read()
                            error = RequestError(
                                request_id=task_id,
                                error_type="HTTPError",
                                message=f"HTTP Error {response.status}",
                                status_code=response.status,
                                headers=dict(response.headers),
                                raw_response=raw_response
                            )
                            self.request_failed.emit(error)
                            return

                        # Log retry information
                        _logger.warning(
                            f"Request {task_id} failed with status {response.status}. "
                            f"Retrying in {retry_delay:.1f}s (attempt {attempt + 1}/{self._max_retries})"
                        )

                        # Wait before retrying
                        await asyncio.sleep(retry_delay)

                except aiohttp.ClientError as e:
                    last_status = getattr(e, 'status', None)

                    if attempt >= self._max_retries:
                        error = RequestError(
                            request_id=task_id,
                            error_type=e.__class__.__name__,
                            message=str(e),
                            status_code=last_status,
                            headers=getattr(e, 'headers', None)
                        )
                        self.request_failed.emit(error)
                        return

                    retry_delay = self._retry_backoff * (2 ** attempt)
                    _logger.warning(
                        f"Request {task_id} failed with {type(e).__name__}. "
                        f"Retrying in {retry_delay:.1f}s (attempt {attempt + 1}/{self._max_retries})"
                    )
                    await asyncio.sleep(retry_delay)

                except asyncio.CancelledError:
                    self.request_cancelled.emit(task_id)
                    _logger.info(f"Request {task_id} was cancelled")
                    raise

                except Exception as e:
                    error = RequestError(
                        request_id=task_id,
                        error_type=type(e).__name__,
                        message=str(e),
                        status_code=last_status
                    )
                    self.request_failed.emit(error)
                    _logger.error(
                        f"Unexpected error in request {task_id}",
                        exc_info=True,
                        extra={"url": url, "method": method, "attempt": attempt}
                    )
                    return
        except asyncio.CancelledError:
            _logger.info(f"Request {task_id} was cancelled during setup")
            self.request_cancelled.emit(task_id)
        except Exception as e:
            error = RequestError(
                request_id=task_id,
                error_type=type(e).__name__,
                message=str(e),
                status_code=last_status
            )
            self.request_failed.emit(error)
            _logger.error(
                f"Critical error processing request {task_id}",
                exc_info=True,
                extra={"url": url, "method": method}
            )

    async def _ensure_session(self) -> None:
        """Ensure we have an active client session."""
        try:
            if self._session is None or self._session.closed:
                timeout = aiohttp.ClientTimeout(
                    total=self._timeout,
                    connect=self._timeout_connect,
                    sock_read=self._timeout,
                    sock_connect=self._timeout_connect
                )

                for attempt in range(self._max_retries + 1):
                    try:
                        self._session = aiohttp.ClientSession(
                            connector=aiohttp.TCPConnector(
                                ssl=self._verify_ssl,
                                limit=self._limit,
                                limit_per_host=self._limit_per_host,
                                force_close=True
                            ),
                            timeout=timeout,
                            trust_env=True
                        )
                        _logger.info("Created new client session")
                        return
                    except Exception as e:
                        if attempt >= self._max_retries:
                            _logger.error("Session creation failed after retries", exc_info=True)
                            raise
                        delay = self._retry_backoff * (2 ** attempt)
                        await asyncio.sleep(delay)
        except asyncio.CancelledError:
            _logger.info("Session creation cancelled")
            raise

    async def _process_response(
            self,
            response: aiohttp.ClientResponse,
            response_type: ResponseType
    ) -> str | dict | bytes | int:
        """Process response content according to specified type."""
        if response_type not in get_args(ResponseType):
            raise ValueError(f"Invalid response_type: {response_type}")

        try:
            if response_type == "text":
                return await response.text(encoding="utf-8")

            elif response_type == "json":
                return await response.json()

            elif response_type == "bytes":
                return await response.read()

            elif response_type == "status":
                return response.status

            else:
                raise ValueError(f"Unsupported response type: {response_type}")
        except asyncio.CancelledError:
            _logger.info("Response processing cancelled")
            raise
        except aiohttp.ContentTypeError as e:
            _logger.error(
                "Content type mismatch",
                extra={"headers": dict(response.headers), "status": response.status}
            )
            raise ValueError("Response content type does not match expected format") from e
        except (UnicodeDecodeError, json.JSONDecodeError) as e:
            raw_data = await response.read()
            sample = raw_data[:200].decode('utf-8', errors='replace') if raw_data else None
            _logger.error(
                "JSON decode failed. Status: %d, Sample: %s",
                response.status, sample,
                exc_info=True
            )

            raise ValueError(f"Invalid JSON response: {str(e)}") from e

    async def _close_session(self, session: aiohttp.ClientSession) -> None:
        if session and not session.closed:
            await session.close()
