from __future__ import annotations

import asyncio
import logging
import threading
import uuid
from asyncio import Task
from typing import Coroutine, Any, Callable

_logger = logging.getLogger(__name__)


class AsyncWorker:
    """
    Thread-safe async task executor running in a background thread.
    
    Features:
    - Automatic error logging for unhandled exceptions in tasks
    - Guaranteed event loop initialization
    - Clean shutdown with task cancellation
    - Task tracking with UUID identifiers
    - Individual task cancellation
    
    Usage:
        worker = AsyncWorker()
        task_id = worker.run_task(coro)
        worker.cancel_task(task_id)
        worker.stop()
    """

    def __init__(self) -> None:
        """
        Initialize worker and start event loop thread.
        Raises:
            RuntimeError: If event loop fails to start within timeout.
        """
        self._lock = threading.Lock()
        self._tasks: dict[uuid.UUID, Task] = {}

        self._loop_ready = threading.Event()
        self._stop_requested = False
        self.loop: asyncio.AbstractEventLoop = asyncio.new_event_loop()
        self.thread = threading.Thread(
            target=self._run_loop,
            daemon=True,
            name="AsyncWorkerThread"
        )
        self.thread.start()

        if not self._loop_ready.wait(timeout=1.0):
            raise RuntimeError("Event loop failed to start within timeout")

    def _run_loop(self) -> None:
        """Main event loop execution method."""
        asyncio.set_event_loop(self.loop)
        self._loop_ready.set()

        try:
            self.loop.run_forever()
        finally:
            # Cleanup even if unexpected error occurs
            self.loop.close()

    def run_task(self,
                 coro_func: Callable[..., Coroutine],
                 *args: Any,
                 **kwargs: Any
                 ) -> uuid.UUID | None:
        """
        Schedule a coroutine for execution.
        :param coro_func: Coroutine to execute
        :return: Task ID if task was scheduled, None if worker is stopping
        """

        with self._lock:
            if self._stop_requested:
                _logger.warning("Worker is shutting down, task rejected")
                return None

        task_id = uuid.uuid4()

        async def wrapped_task() -> None:
            try:
                with self._lock:
                    self._tasks[task_id] = asyncio.current_task()
                await coro_func(task_id, *args, **kwargs)
            except asyncio.CancelledError:
                _logger.error("Unhandled asyncio.CancelledError in async task")
            except BaseException as e:
                _logger.error(f"Unhandled error in async task: {e}")
            finally:
                with self._lock:
                    # noinspection PyAsyncCall
                    self._tasks.pop(task_id)

        asyncio.run_coroutine_threadsafe(wrapped_task(), self.loop)

        return task_id

    def cancel_task(self, task_id: uuid.UUID) -> bool:
        """
        Attempt to cancel a scheduled task.
        :param task_id: UUID of the task to cancel
        :return: True if cancellation was requested, False if task not found
        """
        with self._lock:
            task = self._tasks.get(task_id)
            if task is None:
                return False

            task.cancel()
        return True

    def stop(self,
             coro_stop: Callable[..., Coroutine] | None = None,
             *args: Any,
             **kwargs: Any) -> None:
        """
        Initiate graceful shutdown.
        Cancels all pending tasks and stops the event loop.
        """
        with self._lock:
            if self._stop_requested or not hasattr(self, 'loop') or not self.loop.is_running():
                return

            self._stop_requested = True

        try:
            asyncio.run_coroutine_threadsafe(self._async_stop(coro_stop, *args, **kwargs), self.loop)
        except RuntimeError as e:
            _logger.warning(f"Failed to schedule shutdown coroutine: {e}")

    async def _async_stop(self,
                          coro_stop: Callable[..., Coroutine] | None = None,
                          *args: Any,
                          **kwargs: Any) -> None:
        """Async shutdown procedure."""
        # Cancel all tasks except current
        try:
            tasks_to_cancel = [t for t in asyncio.all_tasks(self.loop)
                               if t is not asyncio.current_task()]

            for task in tasks_to_cancel:
                task.cancel()

            # Allow tasks to handle cancellation
            await asyncio.gather(*tasks_to_cancel, return_exceptions=True)
        except BaseException as e:
            _logger.error(f"An error occurred while cancelling coroutines.: {e}")

        try:
            if coro_stop:
                coroutine = coro_stop(*args, **kwargs)
                if asyncio.iscoroutine(coroutine):
                    await coroutine
                else:
                    _logger.warning("Provided stop handler is not a coroutine")
        except BaseException as e:
            _logger.error(f"An error occurred while cancelling the stop coroutine: {e}")

        # Final stop
        self.loop.stop()

    def get_active_task_count(self) -> int:
        """
        Get number of currently active tasks.
        :return: Count of active tasks
        """
        with self._lock:
            return len(self._tasks)

    def is_running(self) -> bool:
        """
        Check if worker is active.
        :return: True if worker accepts new tasks
        """
        with self._lock:
            return not self._stop_requested

    def __del__(self) -> None:
        """Safety shutdown on garbage collection."""
        if hasattr(self, 'loop') and self.loop.is_running():
            self.stop()
