from collections.abc import Mapping
from typing import Dict, Iterator, Optional

from PySide2.QtCore import QObject, Signal
from axipy import MapTool


class MetaToolManager(type(QObject), type(Mapping)):
    pass


class BaseToolManager(QObject, Mapping, metaclass=MetaToolManager):
    """
    Base tool management system with dictionary-like interface.
    Supports QT signals for adding and removing tools.
    """

    tool_added = Signal(str, MapTool)  # (tool_id: str, tool: MapTool)
    tool_removed = Signal(str, MapTool)  # (tool_id: str, tool: MapTool)
    tools_cleared = Signal()  # Emitted after clear()

    def __init__(self) -> None:
        super().__init__()
        self._tools: Dict[str, MapTool] = {}

    def register_tool(self, tool_id: str, tool: MapTool) -> None:
        """Register a new tool."""
        if tool_id in self._tools:
            raise KeyError(f"Tool '{tool_id}' already registered")

        self._tools[tool_id] = tool
        self.tool_added.emit(tool_id, tool)

    def unregister_tool(self, tool_id: str) -> None:
        """Safely remove a tool."""
        tool = self._tools.get(tool_id)
        if not tool:
            raise KeyError(f"Tool '{tool_id}' not found")

        del self._tools[tool_id]
        self.tool_removed.emit(tool_id, tool)

    def clear(self) -> None:
        """Clear all tools."""
        for tool_id in list(self._tools.keys()):
            self.unregister_tool(tool_id)
        self.tools_cleared.emit()

    # Dictionary interface implementation
    def __getitem__(self, key: str) -> MapTool:
        return self._tools[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._tools)

    def __len__(self) -> int:
        return len(self._tools)

    def __contains__(self, key: str) -> bool:
        return key in self._tools

    def get(self, key: str, default: Optional[MapTool] = None) -> Optional[MapTool]:
        return self._tools.get(key, default)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(tools={list(self._tools.keys())})"

    def __str__(self) -> str:
        return f"Tool Manager ({len(self)} tools)"
