from __future__ import annotations

import logging
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Generator

import axipy

from .. import config
import uuid
from .map_service import map_service
from .wms_params import WMSParams
from .web_service import web_service

LayerID = str
LayerName = str
ObjectKey = str


@dataclass(frozen=True)
class LayerEntry:
    LayerID: str
    LayerName: str


@dataclass
class ObjectData:
    label: str
    properties: dict
    geometry: dict | axipy.Geometry


class DataModel:
    def __init__(self, all_layers: dict[LayerID, LayerName], search_themes: dict[int, str]):
        """
        Инициализируем модель
        :param all_layers: Словарь всех доступных слоёв в формате {WMS_layer_id: layer_name}
        :param search_themes: Словарь тем для поиска
        """
        self._all_layers: dict[LayerID, LayerName] = all_layers

        self._search_mode: bool = False

        self._layers: list[LayerEntry] = []
        self._current_layer_index: int = -1

        self._objects: dict[int, ObjectData] = {}
        self._current_object_id: int = -1

        self._wms_params: WMSParams | None = None
        self._active_request_id: uuid.UUID | None = None

        self._update_object_blocked: bool = False

        self._search_themes = search_themes
        self._current_theme_id: int = -1
        self._search_text: str = ''

        self._logger = logging.getLogger(self.__class__.__name__)

    @property
    def search_mode(self) -> bool:
        return self._search_mode

    @search_mode.setter
    def search_mode(self, new_mode) -> None:
        self._search_mode = new_mode

    @property
    def layers(self) -> list[LayerEntry]:
        """
        Список доступных слоёв
        :return:  Список кортежей в формате (WMS-идентификатор, название_слоя)
        """
        return self._layers

    @layers.setter
    def layers(self, new_layers: list[LayerEntry]) -> None:
        if new_layers == self._layers:
            return

        previous_selection = (
            self._layers[self._current_layer_index]
            if 0 <= self._current_layer_index < len(self._layers)
            else None
        )

        self._layers = new_layers
        self._adjust_layer_selection(previous_selection)

    def _adjust_layer_selection(self, previous: LayerEntry | None) -> None:
        """Update current layer index after layers change."""
        """
        Обновляет индекс текущего слоя после изменения списка слоёв.
        :param previous: Предыдущие значения (WMS-идентификатор, название) выбранного слоя
        """
        if not self._layers:
            self.current_layer_index = -1
            return

        # Try to preserve selection by WMS ID
        if previous:
            new_index = next(
                (i for i, layer in enumerate(self._layers)
                 if layer.LayerID == previous.LayerID),
                -1
            )
            if new_index != -1:
                self.current_layer_index = new_index
                return
        # Fallback to first item
        self.current_layer_index = 0

    @property
    def current_layer_index(self) -> int:
        """Индекс текущего слоя в списке layers (-1 если список пуст)."""
        return self._current_layer_index

    @current_layer_index.setter
    def current_layer_index(self, index: int) -> None:
        if index < -1 or index >= len(self._layers):
            raise ValueError(f"Invalid layer index: {index}")

        if self._current_layer_index == index:
            return

        self._current_layer_index = index
        self._refresh_objects()

    @contextmanager
    def _suppress_object_updates(self) -> Generator[None, None, None]:
        """Контекстный менеджер для временной блокировки обновления объектов."""
        original_state = self._update_object_blocked
        self._update_object_blocked = True
        try:
            yield
        finally:
            self._update_object_blocked = original_state

    def _refresh_objects(self) -> None:
        """Вызываем обновление списка объектов для текущего слоя и WMS-параметров."""
        if self._update_object_blocked:
            return

        if (self._current_layer_index < 0
                or self._current_layer_index >= len(self._layers)
                or self._wms_params is None):
            self.objects = {}
            return

        try:
            layer_id = self._layers[self._current_layer_index].LayerID
            params = self._wms_params

            self._active_request_id = web_service.begin_get_feature_info(
                previous_request_id=self._active_request_id,
                layer_id=layer_id,
                bbox=params.scene_rect,
                map_width=params.device_rect.width,
                map_height=params.device_rect.height,
                i=params.pnt.x,
                j=params.pnt.y,
                feature_count=config.REQUEST_FEATURE_COUNT
            )
        except Exception as e:
            self._logger.error("Failed to fetch feature info: %s", e, exc_info=True)
            self.objects = {}

    @property
    def objects(self) -> dict[int, ObjectData]:
        """Доступные объекты"""
        return self._objects

    @objects.setter
    def objects(self, new_objects: dict[int, ObjectData]) -> None:
        if self._objects.keys() == new_objects.keys():
            return

        self._objects = new_objects
        self._adjust_object_selection()

    def _adjust_object_selection(self) -> None:
        """Обновляет текущий выбранный объект после изменения списка доступных объектов."""
        if not self._objects:
            self.current_object_id = -1
            return

        if self._current_object_id not in self._objects:
            self.current_object_id = next(iter(self._objects.keys()))

    @property
    def current_object_id(self) -> int:
        """
        Текущий выбранный объект.
        :return: id объекта из словаря объектов.
        """
        return self._current_object_id

    @current_object_id.setter
    def current_object_id(self, new_id: int) -> None:
        if new_id == self._current_object_id:
            return

        if new_id >= 0 and new_id not in self._objects:
            raise ValueError(f"Invalid object id: {new_id}")

        self._current_object_id = new_id

    @property
    def wms_params(self) -> WMSParams | None:
        """Текущие параметры WMS-запроса"""
        return self._wms_params

    @wms_params.setter
    def wms_params(self, params: WMSParams | None) -> None:
        if params == self._wms_params:
            return

        # noinspection PyArgumentList
        with self._suppress_object_updates():
            if params:
                try:
                    layers = map_service.get_layers(params.id_mapview)
                    new_layers = list(LayerEntry(layer[1], layer[0]) for layer in layers)
                    if new_layers:
                        self.layers = new_layers
                        self._wms_params = params
                    else:
                        self.layers = []
                        self._wms_params = None
                except Exception as e:
                    self._logger.error("Failed to update WMS parameters: %s", e)
                    self._wms_params = None
                    self.layers = []
            else:
                self._wms_params = None
                self.layers = []

        self._refresh_objects()


    @property
    def active_request_id(self) -> uuid.UUID | None:
        """ID текущего WMS запроса."""
        return self._active_request_id

    @active_request_id.setter
    def active_request_id(self, new_request_id: uuid.UUID):
        self._active_request_id = new_request_id

    @property
    def themes(self) -> dict[int, str]:
        """Вид объектов для поиска"""
        return self._search_themes

    @themes.setter
    def themes(self, new_themes: dict[int, str]) -> None:
        self._search_themes = new_themes

    @property
    def current_theme_id(self) -> int:
        """id текущей темы"""
        return self._current_theme_id

    @current_theme_id.setter
    def current_theme_id(self, new_theme_id: int) -> None:
        if new_theme_id == self._current_theme_id:
            return

        if new_theme_id >= 0 and new_theme_id not in self._search_themes:
            raise ValueError(f"Invalid theme id: {new_theme_id}")

        self._current_theme_id = new_theme_id

    @property
    def search_text(self) -> str:
        return self._search_text

    @search_text.setter
    def search_text(self, new_search_text: str):
        self._search_text = new_search_text

    def reset(self) -> None:
        """Очистить модель."""
        self._layers = []
        self._current_layer_index = -1
        self._objects = {}
        self._current_object_id = -1
        self._wms_params = None
        self._active_request_id = None
