from typing import Any, Dict, List, Optional, Union, Sequence

from PySide2.QtCore import Qt, QSize, QTimer
from PySide2.QtGui import QTextDocument
from PySide2.QtWidgets import QTreeWidget, QTreeWidgetItem, QStyledItemDelegate, QStyle, QApplication


class JsonTreeWidget(QTreeWidget):
    KEY_COLUMN = 0
    VALUE_COLUMN = 1

    def __init__(self, parent: Optional[Any] = None) -> None:
        super().__init__(parent)
        self.setHeaderLabels(["Key", "Value"])
        self.setColumnCount(2)

        self.setWordWrap(True)
        self.setUniformRowHeights(False)
        self.setAnimated(False)
        self.setItemDelegate(WordWrapDelegate(self))

        self.header().sectionResized.connect(self._on_section_resized)

        self._resize_timer = QTimer()
        self._resize_timer.setSingleShot(True)
        self._resize_timer.timeout.connect(self.force_resize_rows)

    def _on_section_resized(self, logical_index: int, old_size: int, new_size: int) -> None:
        """Handle column resize events to update row heights.

        Args:
            logical_index: Index of the resized column
            old_size: Previous width of the column
            new_size: New width of the column
        """
        if logical_index == self.VALUE_COLUMN and old_size != new_size:
            self._resize_timer.start(100)

    def force_resize_rows(self) -> None:
        """Force update of all row heights."""
        self.doItemsLayout()

    def load_json(self, data: Union[Dict, List, Sequence]) -> None:
        """Load JSON data into the tree widget.

        Args:
            data: JSON data to display (dict or list)

        Raises:
            TypeError: If data is not a dict or list
        """
        if not isinstance(data, (dict, list)):
            raise TypeError("Expected dict or list for JSON data")

        self.clear()
        self._add_tree_items(self, data)

    def update_json(self, data: Union[Dict, List, Sequence],
                    preserve_state: bool = True) -> None:
        """Update JSON data while optionally preserving expanded state.

        Args:
            data: New JSON data to display
            preserve_state: Whether to preserve expanded items and selection
        """
        if not data:
            self.clear()
            return

        expanded_items = self._get_expanded_items(self) if preserve_state else []

        selected_text = (self.currentItem().text(self.KEY_COLUMN)
                         if (preserve_state and self.currentItem()
                             and self._is_item_visible(self.currentItem()))
                         else None)

        self.clear()
        self._add_tree_items(self, data)

        if preserve_state:
            self._restore_expanded_items(self, expanded_items)
            if selected_text:
                self._restore_selected_item(self, selected_text)

    def _create_tree_item(self, parent: Union[QTreeWidget, QTreeWidgetItem],
                          key: str = "") -> QTreeWidgetItem:
        """Helper method to create a new tree item."""
        item = QTreeWidgetItem(parent)
        item.setText(self.KEY_COLUMN, key)
        return item

    def _add_tree_items(self, parent: Union[QTreeWidget, QTreeWidgetItem],
                        data: Union[Dict, List, Any]) -> None:
        """Recursively add items to the tree based on JSON data."""
        if isinstance(data, dict):
            for key, value in data.items():
                item = self._create_tree_item(parent, str(key))
                self._add_value_to_item(item, value)
        elif isinstance(data, (list, tuple)):
            for index, value in enumerate(data):
                item = self._create_tree_item(parent, str(index))
                self._add_value_to_item(item, value)
        else:
            item = self._create_tree_item(parent)
            item.setText(self.VALUE_COLUMN, str(data))
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)

    def _add_value_to_item(self, item: QTreeWidgetItem, value: Any) -> None:
        """Add value to the item's value column."""
        if isinstance(value, (dict, list, tuple)):
            self._add_tree_items(item, value)
        else:
            item.setText(self.VALUE_COLUMN, str(value))

    def _get_expanded_items(self, parent: Union[QTreeWidgetItem, QTreeWidget],
                            current_path: str = "",
                            expanded_paths: Optional[List[str]] = None) -> List[str]:
        """Get paths of all expanded items in the tree."""
        expanded_paths = expanded_paths or []

        count = parent.childCount() if isinstance(parent, QTreeWidgetItem) else parent.topLevelItemCount()

        for i in range(count):
            item = parent.child(i) if isinstance(parent, QTreeWidgetItem) else parent.topLevelItem(i)
            item_path = f"{current_path}/{item.text(self.KEY_COLUMN)}" if current_path else item.text(self.KEY_COLUMN)

            if item.isExpanded():
                expanded_paths.append(item_path)
                self._get_expanded_items(item, item_path, expanded_paths)

        return expanded_paths

    def _restore_expanded_items(self, parent: Union[QTreeWidgetItem, QTreeWidget],
                                expanded_paths: List[str], current_path: str = "") -> None:
        """Restore expanded state of items based on their paths."""
        count = parent.childCount() if isinstance(parent, QTreeWidgetItem) else parent.topLevelItemCount()

        for i in range(count):
            item = parent.child(i) if isinstance(parent, QTreeWidgetItem) else parent.topLevelItem(i)
            item_path = f"{current_path}/{item.text(self.KEY_COLUMN)}" if current_path else item.text(self.KEY_COLUMN)

            if item_path in expanded_paths:
                item.setExpanded(True)

            self._restore_expanded_items(item, expanded_paths, item_path)

    def _restore_selected_item(self, parent: Union[QTreeWidget, QTreeWidgetItem],
                               target_text: str) -> bool:
        """Restore selection to the item with matching key text."""
        count = parent.childCount() if isinstance(parent, QTreeWidgetItem) else parent.topLevelItemCount()
        for i in range(count):
            item = parent.child(i) if isinstance(parent, QTreeWidgetItem) else parent.topLevelItem(i)
            if item.text(self.KEY_COLUMN) == target_text:
                self.setCurrentItem(item)
                return True
            if self._restore_selected_item(item, target_text):
                return True
        return False

    def _is_item_visible(self, item: QTreeWidgetItem) -> bool:
        """Check if the item is currently visible in the viewport."""
        widget = item.treeWidget()
        if not widget:
            return False
        index = widget.indexFromItem(item)
        rect = widget.visualRect(index)
        return not rect.isNull() and rect.intersects(widget.viewport().rect())


class WordWrapDelegate(QStyledItemDelegate):
    """Delegate for word-wrapping text in the value column."""

    def __init__(self, tree_widget: JsonTreeWidget):
        super().__init__()
        self.tree_widget = tree_widget

    def _create_text_document(self, text: str, width: float) -> QTextDocument:
        """Create a text document for measuring wrapped text."""
        doc = QTextDocument()
        doc.setDocumentMargin(0)
        doc.setPlainText(text.strip())
        doc.setTextWidth(width)
        return doc

    def paint(self, painter, option, index) -> None:
        """Custom paint method for wrapped text."""
        if index.column() == JsonTreeWidget.VALUE_COLUMN:
            s = QApplication.style()
            s.drawControl(QStyle.CE_ItemViewItem, option, painter, option.widget)

            text = index.data(Qt.DisplayRole)
            if text:
                doc = self._create_text_document(text, option.rect.width())

                painter.save()
                painter.translate(option.rect.topLeft())
                doc.drawContents(painter)
                painter.restore()
        else:
            super().paint(painter, option, index)

    def sizeHint(self, option, index) -> QSize:
        """Calculate size hint considering text wrapping."""
        if index.column() == JsonTreeWidget.VALUE_COLUMN:
            text: str = index.data(Qt.DisplayRole)
            if text:
                width = self.tree_widget.columnWidth(index.column())
                doc = self._create_text_document(text, width)
                height = doc.documentLayout().documentSize().height()
                return QSize(int(doc.idealWidth()), int(height))
        return super().sizeHint(option, index)