from __future__ import annotations

import math
import re
from typing import cast, Final

import axipy

from .. import config
from .singleton import singleton
from .wms_params import WMSParams


@singleton
class MapService:
    _pattern: Final[re.Pattern] = re.compile(r'/(\d+)/wms\?SERVICE=WMS$')
    _schema: Final[axipy.Schema] = axipy.Schema(
        axipy.Attribute.string('id', 255),
        coordsystem=config.SRS)

    def __init__(self) -> None:
        return

    def get_id_mapview(self, mapview: axipy.MapView) -> str | None:
        """
        Получает ID MapView.

        :param mapview: MapView.
        :return: ID MapView или None, если объект некорректен.
        """
        if isinstance(mapview, axipy.MapView):
            return mapview.map.cosmetic.data_object.name

    def get_mapview(self, id_mapview: str) -> axipy.MapView | None:
        """
        Находит MapView по ID.

        :param id_mapview: ID MapView.
        :return: MapView или None, если не найден.
        """
        for mapview in axipy.view_manager.mapviews:
            if self.get_id_mapview(mapview) == id_mapview:
                return mapview
        return None

    def get_id_layer(self, layer: axipy.Layer) -> str | None:
        """
        Получает идентификатор WMS слоя.

        :param layer: Слой.
        :return: ID слоя или None, если слой не WMS.
        """
        if (
                isinstance(layer, axipy.Layer)
                and layer.data_object.provider == 'WmsDataProvider'
                and isinstance(layer.data_object.properties, dict)
        ):
            url = layer.data_object.properties.get('url', '')
            match = self._pattern.search(url)
            if match:
                return match.group(1)
        return None

    def get_layers(self, id_mapview: str) -> list[tuple[str, str]]:
        """
        Получает список слоев, с которых можно получить информацию.

        :param id_mapview: ID MapView
        :return: Список кортежей с названием слоя и его идентификатором или None, если MapView не найден.
        """

        def _get_sources(layers: axipy.ListLayers | axipy.Layer) -> list[tuple[str, str]]:
            sources = []
            for layer in layers:
                if isinstance(layer, axipy.ListLayers) and layer.visible:
                    sources.extend(_get_sources(layer))
                elif isinstance(layer, axipy.RasterLayer) and layer.visible:
                    id_layer = self.get_id_layer(layer)
                    if id_layer:
                        sources.append((layer.title, id_layer))
            return sources

        mapview = self.get_mapview(id_mapview)
        if mapview:
            return _get_sources(mapview.map.layers)
        return []

    def get_wms_params(self,
                       pnt: axipy.Pnt,
                       mapview: axipy.MapView,
                       coordsys: axipy.CoordSystem) -> WMSParams:

        if coordsys.is_equal(mapview.coordsystem, True):
            return WMSParams(self.get_id_mapview(mapview), mapview.device_rect, pnt, mapview.scene_rect)
        else:
            device_len_new = min(mapview.device_rect.width, mapview.device_rect.height)
            device_rect_new = axipy.Rect(0, 0, device_len_new, device_len_new)
            pnt_new = axipy.Pnt(device_len_new // 2, device_len_new // 2)

            center_tmp = axipy.Pnt(*mapview.device_to_scene_transform.map(pnt.x, pnt.y))
            scene_len_tmp = min(mapview.scene_rect.width, mapview.scene_rect.height)
            corners_tmp = (axipy.Pnt(corner[0], corner[1])
                           for corner in
                           [(center_tmp.x - scene_len_tmp / 2, center_tmp.y - scene_len_tmp / 2),
                            (center_tmp.x + scene_len_tmp / 2, center_tmp.y - scene_len_tmp / 2),
                            (center_tmp.x - scene_len_tmp / 2, center_tmp.y + scene_len_tmp / 2),
                            (center_tmp.x + scene_len_tmp / 2, center_tmp.y + scene_len_tmp / 2)])

            transformer = axipy.CoordTransformer(mapview.coordsystem, coordsys)
            center_new: axipy.Pnt = transformer.transform(center_tmp)

            scene_len_new = 2 / math.sqrt(2) * min(
                axipy.Line.distance_by_points(center_new, corner_new)[0] for corner_new in
                (cast(axipy.Pnt, transformer.transform(corner_tmp)) for corner_tmp in corners_tmp))

            scene_rect_new = axipy.Rect(center_new.x - scene_len_new / 2,
                                        center_new.y - scene_len_new / 2,
                                        center_new.x + scene_len_new / 2,
                                        center_new.y + scene_len_new / 2)

            return WMSParams(self.get_id_mapview(mapview), device_rect_new, pnt_new, scene_rect_new)

    def copy_to_clipboard(self, id_mapview: str, geom: axipy.Geometry, geom_id: str) -> bool:
        mv = self.get_mapview(id_mapview)
        if not mv:
            return False

        definition = {
            'src': '',
            'schema': self._schema,
            'hidden': True
        }
        with axipy.provider_manager.create_open(definition) as table:
            table.insert(axipy.Feature({'id': geom_id}, geom))
            lr = axipy.Layer.create(table)
            lr.title = config.NAME
            mv.map.layers.append(lr)
            axipy.selection_manager.add(table, -1)
            axipy.action_manager.activate('Copy')

        return True


map_service = MapService()
