from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Final

import axipy
from PySide2.QtCore import Slot, QSize, QObject
from PySide2.QtGui import QIcon

from .. import config
from .data_model import DataModel
from .nspd_presenter import NSPDPresenter
from .tool import Tool
from .tool_manager import ToolManager
from .widget import Widget

ICON_PATH: Final[Path] = config.PATH / "images"
ICON_NAME: Final[str] = "nspd_info.png"
ICON_SIZES: Final[tuple[int, ...]] = (16, 24, 32)


class NSPDInfo(QObject):
    def __init__(self, plugin: axipy.Plugin) -> None:
        super().__init__()
        layers = self.load_layers()
        themes = self.load_search_themes()

        self._model = DataModel(layers, themes)
        self._widget = Widget()
        self._tools = ToolManager()
        self._presenter = NSPDPresenter(self._model, self._widget, self._tools)

        self._tool_panel = axipy.ActiveToolPanel().make_custom(
            title=config.NAME,
            observer_id=axipy.ObserverManager.ActiveMapView,
            widget=self._widget)

        self._tool_button = plugin.create_tool(
            title=config.NAME,
            tooltip=config.DESCRIPTION,
            icon=self._create_icon(),
            enable_on=axipy.ObserverManager.ActiveMapView,
            on_click=self._on_tool_button_click
        )
        self._tool_button.action.toggled.connect(self._on_tool_button_toggled)

        position = axipy.Position(plugin.tr(config.BUTTON_TAB), plugin.tr(config.BUTTON_GROUP))
        position.add(self._tool_button)

    def unload(self) -> None:
        self._presenter.unload()
        self._tool_button.remove()
        self._widget.unload()
        self._widget.deleteLater()
        self._widget.setParent(None)
        del self._widget

    @Slot(bool)
    def _on_tool_button_toggled(self, checked: bool) -> None:
        if checked:
            self._tool_panel.activate()
        else:
            self._presenter.clear()
            self._tool_panel.deactivate()

    def _on_tool_button_click(self) -> Tool:
        """
        Создаём инструмент.
        """
        return Tool(self._tools)

    def load_layers(self) -> dict[str, str]:
        """
        Загружаем словарь слоёв ключи - layerId, значения названия
        :return:
        """
        # TODO: Не используются
        path = Path(os.path.dirname(os.path.abspath(__file__))) / 'services'
        layers = dict()
        for file_name in path.rglob('*.json'):
            with open(file_name, 'r', encoding='utf-8') as file:
                data = json.load(file)
                if isinstance(data, dict) and 'layers' in data:
                    file_layers = {lr['layerId']: lr['title'] for lr in data['layers'] if lr['layerType'] == 'wms'}
            layers.update(file_layers)
        return layers

    def load_search_themes(self) -> dict[int, str]:
        """Виды объектов для поиска"""
        file_name = config.PATH / 'data' / 'search-theme.json'
        with open(file_name, 'r', encoding='utf-8') as file:
            data = json.load(file)
            themes = {theme['id']: theme['name'] for theme in data}
            return themes

    def _create_icon(self) -> QIcon:
        icon = QIcon()
        for size in ICON_SIZES:
            icon_path: str = str(ICON_PATH / f"{size}px" / ICON_NAME)
            icon.addFile(icon_path, QSize(size, size))
        return icon
