from __future__ import annotations

import json
import logging
import math
import uuid
from typing import cast

import axipy
from PySide2.QtCore import QObject, Slot

from .async_http_client import RequestResult, RequestError
from .data_model import DataModel, ObjectData
from .map_service import map_service
from .tool import Tool
from .tool_manager import ToolManager
from .web_service import web_service
from .widget import Widget
from .. import config


class NSPDPresenter(QObject):
    def __init__(self,
                 model: DataModel,
                 widget: Widget,
                 tools: ToolManager):
        super().__init__()

        self._model = model
        self._widget = widget
        self._tools = tools

        self._active_tool: str | None = None

        self._widget.copy_requested.connect(self._on_copy_requested)
        self._widget.layer_changed.connect(self._on_layer_changed)
        self._model.current_theme_id = next(iter(self._model.themes.keys()), -1)
        self._widget.set_search_themes(self._model.themes, self._model.current_theme_id)
        self._widget.search_theme_changed.connect(self._on_search_theme_changed)
        self._widget.search_text_changed.connect(self._on_search_text_changed)
        self._widget.search_requested.connect(self._on_search_requested)

        self._widget.object_changed.connect(self._on_object_changed)

        self._tools.click.connect(self._on_tool_click)
        self._tools.tool_removed.connect(self._on_tool_removed)
        web_service.client.request_started.connect(self._on_client_request_started)
        web_service.client.request_completed.connect(self._on_client_request_completed)
        web_service.client.request_failed.connect(self._on_client_request_failed)

        self._logger = logging.getLogger(self.__class__.__name__)

    @Slot(int)
    def _on_layer_changed(self, index: int) -> None:
        self._model.current_layer_index = index

        self.update_tool()
        self.update_widget()

    @Slot('long long')
    def _on_object_changed(self, key: int) -> None:
        self._model.current_object_id = key

        self.update_tool()
        self.update_widget()

    @Slot(int)
    def _on_theme_id_changed(self, key: int) -> None:
        self._model.current_theme_id = key

    @Slot()
    def _on_copy_requested(self) -> None:
        obj = self._model.objects.get(self._model.current_object_id)
        if obj:
            map_service.copy_to_clipboard(self._model.wms_params.id_mapview, obj.geometry,
                                          self._model.current_object_id)

    @Slot(int)
    def _on_search_theme_changed(self, theme_id: int) -> None:
        self._model.current_theme_id = theme_id

    @Slot(str)
    def _on_search_text_changed(self, search_text: str) -> None:
        self._model.search_text = search_text

    @Slot()
    def _on_search_requested(self) -> None:
        self._model.active_request_id = web_service.begin_search(
            self._model.active_request_id,
            self._model.current_theme_id,
            self._model.search_text)

    @Slot(str, axipy.Pnt)
    def _on_tool_click(self, tool_id: str, pnt: axipy.Pnt) -> None:
        wms_params = map_service.get_wms_params(
            pnt,
            self._tools[tool_id].view,
            config.SRS)

        self._model.wms_params = wms_params

    @Slot(str)
    def _on_tool_removed(self, tool_id: str):
        if tool_id == self._active_tool:
            self._model.layers = []
            self._model.wms_params = None
            self.update_widget()
            web_service.clear()

    @Slot(uuid.UUID)
    def _on_client_request_started(self, request_id: uuid.UUID):
        self._widget.start_waiting()

    @Slot(RequestResult)
    def _on_client_request_completed(self, result: RequestResult):
        if self._model.active_request_id != result.request_id:
            return

        is_search_query = False
        try:
            if not web_service.client.has_active_requests():
                self._widget.stop_waiting()

            if not isinstance(result.data, dict):
                self._model.objects = {}
                return

            features = result.data.get('features', [])

            if not features:
                data = result.data.get('data', {})
                features = data.get('features', {})
                if features:
                    is_search_query = True

            if not features:
                self._model.objects = {}
                return

            features_data: dict[int, any] = {}
            for feature in features:
                try:
                    feature_id = feature.get('id')
                    if not feature_id:
                        raise ValueError('The feature does not contain id.')
                    geometry = feature.get('geometry')
                    if not geometry:
                        raise ValueError(f'The feature id:{feature_id} does not contain geometry.')
                    properties = feature.get('properties')
                    if not properties:
                        raise ValueError(f'The feature id:{feature_id} does not contain properties.')
                    label = properties.get('label')
                    if label:
                        features_data[feature_id] = ObjectData(
                            label=label,
                            properties=properties,
                            geometry=geometry)
                        continue
                    options = properties.get('options')
                    if not options:
                        features_data[feature_id] = ObjectData(
                            label=f'feature_id={feature_id}',
                            properties=properties,
                            geometry=geometry)
                        continue

                    cad_num = options.get('cad_num')
                    if cad_num:
                        features_data[feature_id] = ObjectData(
                            label=cad_num,
                            properties=properties,
                            geometry=geometry)
                        continue

                    cad_number = options.get('cad_number')
                    if cad_number:
                        features_data[feature_id] = ObjectData(
                            label=cad_number,
                            properties=properties,
                            geometry=geometry)
                        continue

                    features_data[feature_id] = ObjectData(
                        label=str(feature_id),
                        properties=properties,
                        geometry=geometry)
                    # raise ValueError(f"Can't determine the title for feature: id:{feature_id}")

                except Exception as e:
                    self._logger.error(f"Error parse request:\n{e}")

            self._model.objects = dict(sorted(features_data.items(), key=self._objects_sort_key))
        except Exception as e:
            self._logger.error(f"Error request_completed:\n{e}")
            self._model.objects = {}
        finally:
            self._model.search_mode = is_search_query
            self.update_tool()
            self.update_widget()

    @Slot(RequestError)
    def _on_client_request_failed(self, error: RequestError) -> None:
        if not web_service.client.has_active_requests():
            self._widget.stop_waiting()
        if error.raw_response:
            axipy.Notifications.push(config.NAME,
                                     f'Произошла ошибка при обращении к НСПД'
                                     f'\n{error.message}'
                                     f'\n{error.raw_response}',
                                     axipy.Notifications.Warning)
        else:
            axipy.Notifications.push(config.NAME,
                                     f'Произошла ошибка при обращении к НСПД'
                                     f'\n{error.message}',
                                     axipy.Notifications.Warning)

    def _objects_sort_key(self, pair: tuple[int, ObjectData]) -> tuple:
        """
        Создает ключ сортировки для меток.
        Группа из нулей считается самой большой.
        :param pair: Кортеж (id_объекта, данные_объекта)
        :return: Ключ сортировки
        """
        return tuple(
            math.inf if group == '0' * len(group) else int(group) if group.isdigit() else group
            for group in pair[1].label.split(':')
        )

    def update_tool(self) -> None:
        if self._active_tool and (
                self._model.wms_params is None or
                self._active_tool != self._model.wms_params.id_mapview):

            old_tool = cast(Tool, self._tools.get(self._active_tool))
            if old_tool:
                old_tool.geometry = None

        if self._model.current_object_id >= 0 and self._model.wms_params:

            self._active_tool = self._model.wms_params.id_mapview
            tool = cast(Tool, self._tools.get(self._active_tool))

            if tool:
                try:
                    geometry = self._model.objects[self._model.current_object_id].geometry
                    if isinstance(geometry, axipy.Geometry):
                        tool.geometry = geometry
                    else:
                        # TODO Если геометрия большая - from_geojson может быть медленно. Можно вызывать асинхронно
                        geom = axipy.Geometry.from_geojson(json.dumps(geometry), config.SRS)
                        self._model.objects[self._model.current_object_id].geometry = geom
                        tool.geometry = geom
                except Exception as e:
                    self._logger.error(f"Error create geometry: {e}")
            else:
                self._active_tool = None

    def update_widget(self) -> None:
        if self._model.search_mode:
            self._widget.set_layers([], -1)
        else:
            self._widget.set_layers(
                [layer.LayerName for layer in self._model.layers],
                self._model.current_layer_index
            )

        self._widget.set_objects(
            {id: obj.label for id, obj in self._model.objects.items()},
            self._model.current_object_id
        )

        current_object = self._model.objects.get(self._model.current_object_id)
        if current_object:
            self._widget.set_data(current_object.properties)
        else:
            self._widget.set_data(None)

    def clear(self) -> None:
        self._model.reset()
        web_service.clear()
        self.update_tool()
        self.update_widget()

    def unload(self) -> None:
        self.clear()
        web_service.unload()
