from __future__ import annotations

import logging

import axipy
from PySide2.QtCore import Qt
from PySide2.QtGui import QPainter, QPaintEvent

from .. import config
from .map_service import map_service
from .tool_manager import ToolManager


class Tool(axipy.MapTool):
    def __init__(self, tools: ToolManager) -> None:
        super().__init__()

        self._tools = tools

        self._id_mapview: str = ""

        self._style = axipy.Style.from_mapinfo(config.SELECT_STYLE)

        self._geometry: axipy.Geometry | None = None
        self._geometry_proj: axipy.Geometry | None = None

        self._logger = logging.getLogger(self.__class__.__name__)

    def load(self) -> None:
        self._id_mapview = map_service.get_id_mapview(self.view)
        self._tools.register_tool(self._id_mapview, self)

    def unload(self) -> None:
        self._tools.unregister_tool(self._id_mapview)

    def mouseReleaseEvent(self, event) -> bool:
        if event.button() != Qt.LeftButton:
            return self.PassEvent

        self._tools.on_click(self._id_mapview, axipy.Pnt(event.x(), event.y()))
        return self.BlockEvent

    def paintEvent(self, event: QPaintEvent, painter: QPainter) -> None:
        if self._geometry:
            try:
                if not self.view.coordsystem.is_equal(self._geometry_proj.coordsystem, True):
                    self._geometry_proj = self._geometry.reproject(self.view.coordsystem)

                geometry = self._geometry_proj.affine_transform(self.view.scene_to_device_transform)

                painter.save()

                painter.setOpacity(config.SELECT_STYLE_OPACITY)
                self._style.draw(geometry, painter)
            except Exception as e:
                self._logger.error(f"Error draw geometry: {e}")
            finally:
                painter.restore()
        super().paintEvent(event, painter)

    @property
    def geometry(self) -> axipy.Geometry | None:
        return self._geometry

    @geometry.setter
    def geometry(self, value: axipy.Geometry | None):
        self._geometry = value
        self._geometry_proj = value
        self.redraw()

    @property
    def id_mapview(self) -> str:
        return self._id_mapview
