from __future__ import annotations

import logging
import uuid
from random import random
from typing import Final

import axipy

from .singleton import singleton
from .. import config
from .async_http_client import AsyncHttpClient


@singleton
class WEBService:
    _layer_template: Final[str] = config.LAYER_TEMPLATE
    _search_template: Final[str] = config.SEARCH_TEMPLATE
    _referer: Final[str] = config.REFERER

    def __init__(self) -> None:
        self._client = AsyncHttpClient(
            max_concurrent=4,
            # timeout=config.CONNECTION_TIMEOUT,
            # timeout_connect=config.CONNECTION_TIMEOUT_CONNECT,
            # limit=config.CONNECTION_LIMIT,
            # limit_per_host=config.CONNECTION_LIMIT_PER_HOST,
            # max_retries=config.REQUEST_MAX_RETRIES,
            # retry_backoff=config.REQUEST_RETRY_BACKOFF,
            verify_ssl=config.CONNECTION_VERIFY_SSL
        )
        self._logger = logging.getLogger(self.__class__.__name__)

    def begin_get_feature_info(self,
                               previous_request_id: uuid.UUID | None,
                               layer_id: str,
                               bbox: axipy.Rect,
                               map_width: int,
                               map_height: int,
                               i: int,
                               j: int,
                               feature_count: int = 1) -> uuid.UUID:
        """
        Асинхронное получение информации о feature с WMS сервера.
        Если предыдущий запрос выполняется, он будет отменён

        :param previous_request_id: Предыдущий идентификатор запроса.
        :param layer_id: ID слоя.
        :param bbox: Ограничивающий прямоугольник (bounding box).
        :param map_width: Ширина карты.
        :param map_height: Высота карты.
        :param i: Координата X на карте.
        :param j: Координата Y на карте.
        :param feature_count: Количество feature для запроса.
        :return: Уникальный идентификатор запроса
        """

        if previous_request_id:
            self._client.cancel_request(previous_request_id)

        base_url = self._layer_template.format(layer_id=layer_id)

        random_value = random()

        # Параметры запроса
        params = {
            "REQUEST": "GetFeatureInfo",
            "QUERY_LAYERS": layer_id,
            "SERVICE": "WMS",
            "VERSION": "1.3.0",
            "FORMAT": "image/png",
            "STYLES": "",
            "TRANSPARENT": "true",
            "LAYERS": layer_id,
            "RANDOM": random_value,
            "INFO_FORMAT": "application/json",
            "FEATURE_COUNT": feature_count,
            "I": int(i),
            "J": int(j),
            "WIDTH": int(map_width),
            "HEIGHT": int(map_height),
            "CRS": "EPSG:3857",
            "BBOX": "{},{},{},{}".format(bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax)
        }

        headers = {
            "referer": self._referer
        }

        return self._client.send_request(
            base_url,
            params=params,
            headers=headers,
            response_type='json'
        )

    def begin_search(self,
                     previous_request_id: uuid.UUID | None,
                     theme_id: int,
                     query: str) -> uuid.UUID:
        """
        Асинхронный поиск
        :param previous_request_id: Предыдущий идентификатор запроса.
        :param theme_id: Код вида объекта для поиска
        :param query: Запрос
        :return: Уникальный идентификатор запроса
        """
        if previous_request_id:
            self._client.cancel_request(previous_request_id)

        base_url = self._search_template
        params = {
            "thematicSearchId": theme_id,
            "query": query}

        headers = {
            "referer": self._referer
        }

        return self._client.send_request(
            base_url,
            params=params,
            headers=headers,
            response_type='json'
        )

    @property
    def client(self) -> AsyncHttpClient:
        return self._client

    def clear(self) -> None:
        try:
            self._client.shutdown()
        except Exception as e:
            self._logger.error(f"Error when clear web_service: {str(e)}")

    def unload(self) -> None:
        try:
            self._client.shutdown()
        except Exception as e:
            self._logger.error(f"Error when unload web_service: {str(e)}")


web_service = WEBService()
