from __future__ import annotations

import axipy
from PySide2.QtCore import Slot, Signal, Qt, QObject
from PySide2.QtGui import QMovie, QIcon, QPixmap, QColor, QFont, QPainter
from PySide2.QtWidgets import QWidget

from .. import config
from .ui_widget import Ui_Widget


class Widget(QWidget, Ui_Widget):
    copy_requested = Signal()
    layer_changed = Signal(int)

    search_theme_changed = Signal(int)
    search_text_changed = Signal(str)
    search_requested = Signal()

    object_changed = Signal('long long')

    def __init__(self, parent: QObject = None) -> None:
        super().__init__(parent)
        self.setupUi(self)
        self._show_search(False)

        self._btn_copy.setIcon(axipy.action_manager.icons['copy'])
        self._btn_show_search.setIcon(axipy.action_manager.icons["search"])

        self._btn_copy.clicked.connect(self._on_copy_clicked)
        self._btn_show_search.clicked.connect(self._on_show_search_clicked)
        self._btn_hide_search.clicked.connect(self._on_hide_search_clicked)

        self._cmb_search_theme.currentIndexChanged.connect(self._on_search_theme_changed)
        self._txt_search.textChanged.connect(self._on_text_changed)
        self._txt_search.editingFinished.connect(self._on_search_edited)
        self._btn_search.clicked.connect(self._on_search_clicked)

        self._cmb_layer.currentIndexChanged.connect(self._on_layer_changed)

        self._cmb_object.currentIndexChanged.connect(self._on_object_changed)

        self._movie = QMovie(str(config.PATH / "images" / "waiting.gif"))
        self._lbl_wait.setMovie(self._movie)
        self._lbl_wait.setVisible(False)

    def create_text_icon(self, text: str = '×', size: int = 16, color: QColor = QColor("black")):
        # Создаем прозрачный pixmap
        pixmap = QPixmap(size, size)
        pixmap.fill(Qt.transparent)

        # Настраиваем painter для отрисовки
        painter = QPainter(pixmap)
        painter.setRenderHint(QPainter.Antialiasing)  # Сглаживание
        painter.setPen(color)

        # Устанавливаем шрифт (можно выбрать другой)
        font = QFont()
        font.setPixelSize(int(size * 0.8))  # Размер символа
        painter.setFont(font)

        # Рисуем Unicode-символ
        painter.drawText(pixmap.rect(), Qt.AlignCenter, text)
        painter.end()

        return QIcon(pixmap)

    def unload(self) -> None:
        self._movie.stop()
        self._lbl_wait.setMovie(None)
        self._movie.deleteLater()

        self._lbl_wait.deleteLater()

    @Slot(int)
    def _on_layer_changed(self, index: int) -> None:
        self.layer_changed.emit(index)

    @Slot(int)
    def _on_object_changed(self, index: int) -> None:
        if index < 0:
            self.object_changed.emit(-1)
        else:
            self.object_changed.emit(self._cmb_object.currentData())

    @Slot(int)
    def _on_search_theme_changed(self, index: int) -> None:
        if index < 0:
            self.search_theme_changed.emit(-1)
        else:
            self.search_theme_changed.emit(self._cmb_search_theme.currentData())

    @Slot(str)
    def _on_text_changed(self, text: str) -> None:
        self._btn_search.setEnabled(bool(text))

    @Slot()
    def _on_search_edited(self) -> None:
        self.search_text_changed.emit(self._txt_search.text())

    @Slot(bool)
    def _on_search_clicked(self, checked: bool):
        if self._txt_search.hasFocus():
            self._txt_search.editingFinished.emit()
        self.search_requested.emit()

    @Slot(bool)
    def _on_copy_clicked(self, checked: bool) -> None:
        self.copy_requested.emit()

    @Slot(bool)
    def _on_show_search_clicked(self, checked: bool) -> None:
        self._show_search(checked)

    @Slot(bool)
    def _on_hide_search_clicked(self, checked: bool) -> None:
        self._btn_show_search.click()

    def _show_search(self, visible: bool) -> None:
        if visible:
            self._grp_search.setVisible(True)
            self._lbl_layer.setVisible(False)
            self._cmb_layer.setVisible(False)
        else:
            self._grp_search.setVisible(False)
            self._lbl_layer.setVisible(True)
            self._cmb_layer.setVisible(True)

    def set_layers(self, layers: list[str], index: int) -> None:
        self._cmb_layer.blockSignals(True)
        try:
            self._cmb_layer.clear()
            self._cmb_layer.addItems(layers)
            if index >= 0:
                self._cmb_layer.setCurrentIndex(index)
        finally:
            self._cmb_layer.blockSignals(False)

        self._cmb_layer.setEnabled(len(layers) > 1)

    def set_objects(self, objects: dict[int, str], current_id: int) -> None:
        self._cmb_object.blockSignals(True)
        try:
            self._cmb_object.clear()
            for key, value in objects.items():
                self._cmb_object.addItem(value, key)

            index = self._cmb_object.findData(current_id)
            if index >= 0:
                self._cmb_object.setCurrentIndex(index)
        finally:
            self._cmb_object.blockSignals(False)

        self._cmb_object.setEnabled(len(objects) > 1)

    def set_search_themes(self, themes: dict[int, str], current_id: int = -1):
        self._cmb_search_theme.blockSignals(True)
        try:
            self._cmb_search_theme.clear()
            for key, value in themes.items():
                self._cmb_search_theme.addItem(value, key)

            index = self._cmb_search_theme.findData(current_id)
            if index >= 0:
                self._cmb_search_theme.setCurrentIndex(index)
        finally:
            self._cmb_search_theme.blockSignals(False)

        self._cmb_search_theme.setEnabled(len(themes) > 1)

    def set_data(self, data: list | dict | None) -> None:
        if data:
            self._tree_property.update_json(data)
        else:
            self._tree_property.update_json(None)

    @Slot()
    def start_waiting(self) -> None:
        if not self._lbl_wait.isVisible():
            self._lbl_wait.setVisible(True)
            self._movie.start()

    @Slot()
    def stop_waiting(self) -> None:
        if self._lbl_wait.isVisible():
            self._movie.stop()
            self._lbl_wait.setVisible(False)
