import json
import logging
import os
from typing import Optional

from PySide2.QtCore import Qt, QObject
from PySide2.QtGui import QIcon
from PySide2.QtWidgets import (
    QDockWidget, QWidget, QVBoxLayout
)
from axipy import (
    Plugin, Position, mainwindow
)

from .nspd_tree_widget import NspdTreeWidget


class NSPDLayers(QObject):
    """
    Инструмент работы с данными НСПД
    """

    def __init__(self, plugin: Plugin) -> None:
        super().__init__()

        self.plugin = plugin
        self.__title = plugin.tr('Карты НСПД')

        self.__icon = QIcon(str(plugin.plugin_dir / 'images' / 'nspd.png'))
        self.__button = plugin.create_action(
            self.__title,
            icon=self.__icon,
            on_click=self.show_widget,
            tooltip=plugin.tr('Каталог национальной системы пространственных данных'))
        position = Position(plugin.tr('Основные'), plugin.tr('Команды'))
        position.add(self.__button)
        self.__dock = None

    def load(self) -> None:
        show_widget = self.plugin.settings.value('ShowWidget', False)
        if (isinstance(show_widget, bool) and show_widget or
                isinstance(show_widget, str) and show_widget == 'true'):
            self.show_widget()

    def __remove_dock(self) -> None:
        if self.__dock is not None:
            self.__dock.close()
            self.__dock = None

    def unload(self) -> None:
        """
        Выгрузка модуля. Удаление ресурсов.
        """
        self.plugin.settings.setValue('ShowWidget', self.__dock is not None)
        self.__remove_dock()
        self.__button.remove()
        self.plugin = None

    def __write_file(self, file_name: str, data: bytes) -> None:
        """
        Сохранение json в файловой системе
        """
        if os.path.isfile(file_name):
            bak_file_name = file_name + '.BAK'
            if os.path.exists(bak_file_name):
                os.remove(bak_file_name)
            os.rename(file_name, bak_file_name)
        f = open(file_name, "wb")
        f.write(data)
        f.close()

    def __retrieve_file(self, file_name: str, url: str) -> None:
        """
        Загрузка перечня с сайта и сохранение в виде json файла 
        """
        import urllib.request as r
        req = r.Request(url, headers={'User-Agent': 'Mozilla/5.0'})
        req.add_header('referer', 'https://nspd.gov.ru/map/')
        try:
            logging.info(f"Получаем: {file_name} по адресу {url}")
            data = r.urlopen(req).read()
            if data:
                self.__write_file(file_name, data)
        except Exception as e:
            logging.critical(f"Ошибка: {e} {type(e).__name__} {e.args}")

    def _json_service_file(self, name: str, url: Optional[str] = None) -> str:
        fn = str(self.plugin.plugin_dir / 'data' / 'services' / name)
        if not os.path.isfile(fn) and url:
            self.__retrieve_file(fn, url)
        return fn

    def _json_file(self, name: str) -> str:
        return str(self.plugin.plugin_dir / 'data' / name)

    def __clear_dock(self) -> None:
        self.__dock = None

    def __create_tree_widget(self, parent: QWidget) -> NspdTreeWidget:
        w = NspdTreeWidget(parent)
        settings = self._json_file('rosreestr.json')
        logging.debug(f'Load settings file: {settings}')
        # Чтение из файла rosreestr.json списка доступных ресурсов
        with open(settings, 'r', encoding='utf-8') as file:
            services = json.load(file)
            # Последовательная загрузка ресурсов в соотвествие с их внутренней структурой                
            for srv in services['services']:
                fn = self._json_service_file(srv['file_name'], srv['url'])
                if os.path.isfile(fn):
                    w.load_file(fn, srv)
                else:
                    logging.critical(f"Ошибка: файл '{fn}' не найден")
        return w

    def show_widget(self) -> None:
        """
        Создание, заполнение и показ древовидной структуры доступных ресурсов.
        """
        if self.__dock is None:
            self.__dock = QDockWidget(self.__title)
            self.__dock.setAttribute(Qt.WA_DeleteOnClose)
            self.__dock.setObjectName("NspdDock")
            widget = QWidget(self.__dock)
            layout = QVBoxLayout()
            tw = self.__create_tree_widget(widget)
            tw.setWindowTitle(self.__title)
            layout.addWidget(tw)
            widget.setLayout(layout)
            self.__dock.setWidget(widget)
            self.__dock.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)
            self.__dock.destroyed.connect(self.__clear_dock)
            mainwindow.add_dock_widget(self.__dock, Qt.RightDockWidgetArea, self.__icon)
        else:
            self.__remove_dock()
